"""Entities for the QwikSwitch API."""

from __future__ import annotations

from http import HTTPStatus
from typing import Any

from .constants import DEVICES, DeviceClass, JsonKeys
from .exceptions import QSResponseParseError
from .utility import ResponseParser


class ApiKeys:
    """API keys for the QwikSwitch API."""

    def __init__(self, read_key: str, read_write_key: str) -> None:
        """
        Initialize an ApiKeys object.

        :param read_key: The API key used for read operations.
        :param read_write_key: The API key used for read-write operations.
        """
        self._read_key = read_key
        self._read_write_key = read_write_key

    @property
    def read_key(self) -> str:
        """
        The API key for read operations (no device control).

        :return: the read API key
        """
        return self._read_key

    @property
    def read_write_key(self) -> str:
        """
        The API key for read-write operations (device control).

        The read-write API key can also be used for read access.

        :return: the read/write API key
        """
        return self._read_write_key

    @classmethod
    def from_resp(cls, resp) -> ApiKeys:  # noqa: ANN001
        """
        Construct an ApiKeys object from JSON data.

        :param resp: The response object to construct the object from
        :return: the ApiKeys object
        :raises QSRequestError: on failure of the response, or validation error.
        """
        if resp.status_code != HTTPStatus.OK:
            ResponseParser.raise_auth_failure(resp)

        json_data = resp.json()

        if (JsonKeys.OK in json_data and json_data[JsonKeys.OK] == 0) or (
            JsonKeys.ERR in json_data
        ):
            ResponseParser.raise_auth_failure(resp)

        return cls(json_data[JsonKeys.READ_KEY], json_data[JsonKeys.READ_WRITE_KEY])


class ControlResult:
    """Result of a control operation on a device."""

    def __init__(self, device_id: str, level: int) -> None:
        """
        Initialize a ControlResult object.

        :param device_id: the unique device identifier
        :param level: The final level to which the device was set
        """
        self._device_id = device_id
        self._level = level

    @property
    def device_id(self) -> str:
        """
        The unique identifier of the device.

        :return: The unique identifier of the device
        """
        return self._device_id

    @property
    def level(self) -> int:
        """
        The level to which the device was set.

        :return: The level to which the device was set
        """
        return self._level

    @classmethod
    def from_resp(cls, resp) -> ControlResult:  # noqa: ANN001
        """
        Construct a ControlResult object from JSON data.

        :param resp: The response object to construct the object from
        :return: A ControlResult object
        :raises QSRequestError: on failure of the response, or validation error
        """
        if resp.status_code != HTTPStatus.OK:
            ResponseParser.raise_request_error(resp)

        json_data = resp.json()

        if (JsonKeys.SUCCESS in json_data and not json_data[JsonKeys.SUCCESS]) or (
            JsonKeys.ERROR in json_data
        ):
            ResponseParser.raise_request_error(resp)

        return cls(json_data[JsonKeys.DEVICE], json_data[JsonKeys.LEVEL])


class DeviceStatus:
    """Status of a device."""

    def __init__(  # noqa: PLR0913
        self,
        device_id: str,
        device_type: str,
        firmware: str,
        epoch: int,
        rssi: int,
        value: int,
    ) -> None:
        """
        Initialize a DeviceStatus object.

        :param device_id: the unique device identifier
        :param device_type: the type of device
        :param firmware: the version of the device firmware
        :param epoch: the epoch time of the last status update
        :param rssi: the signal strength of the device
        :param value: the current value of the device
        """
        self._device_id = device_id
        self._device_type = device_type
        self._firmware = firmware
        self._epoch = epoch
        self._rssi = rssi
        self._value = value

    @property
    def device_id(self) -> str:
        """
        The unique identifier of the device.

        :return: the unique identifier of the device
        """
        return self._device_id

    @property
    def device_type(self) -> str:
        """
        The type of device.

        :return: the type of device
        """
        return self._device_type

    @property
    def firmware(self) -> str:
        """
        The version of the device firmware.

        :return: The version of the device firmware
        """
        return self._firmware

    @property
    def epoch(self) -> int:
        """
        The epoch time of the last status update.

        :return: The epoch time of the last status update
        """
        return self._epoch

    @property
    def rssi(self) -> int:
        """
        The signal strength of the device.

        :return: The signal strength of the device (percentage, 0 - 100)
        """
        return self._rssi

    @property
    def value(self) -> int:
        """
        The current value of the device.

        :return: The current value of the device. [0=off 1-100=on]
        """
        return self._value

    @property
    def device_class(self) -> DeviceClass | Any:
        """
        The class of the device.

        :return: The class of the device
        """
        if self._device_type in DEVICES:
            return DEVICES[self._device_type]
        return DeviceClass.unknown

    @classmethod
    def from_json(cls, json_data) -> DeviceStatus:  # noqa: ANN001
        """
        Construct a DeviceStatus object from JSON data.

        :param json_data: The JSON data to construct the object from
        :return: A DeviceStatus object
        :raises QSRequestError: on validation error
        """
        if len(json_data) > 1:
            msg = f"Expected one device status, got {len(json_data)}"
            raise QSResponseParseError(msg)

        device_id = next(iter(json_data))  # Only expecting one key
        state_json_data = json_data[device_id]

        rssi = int(state_json_data[JsonKeys.RSSI].replace("%", ""))
        return cls(
            device_id,
            state_json_data[JsonKeys.TYPE],
            state_json_data[JsonKeys.FIRMWARE],
            state_json_data[JsonKeys.EPOCH],
            rssi,
            state_json_data[JsonKeys.VALUE],
        )


class DeviceStatuses:
    """Statuses of multiple devices."""

    def __init__(self, statuses: list[DeviceStatus]) -> None:
        """
        Initialize a DeviceStatuses object.

        :param statuses: The list of device statuses
        """
        self._statuses = statuses

    @property
    def statuses(self) -> list[DeviceStatus]:
        """
        The list of device statuses.

        :return: The list of device statuses
        """
        return self._statuses

    @classmethod
    def from_resp(cls, resp) -> DeviceStatuses:  # noqa: ANN001
        """
        Construct a DeviceStatuses object from JSON data.

        :param json_data: The JSON data to construct the object from
        :return: A DeviceStatuses object
        :raises QSRequestError: on validation error
        """
        if resp.status_code != HTTPStatus.OK:
            ResponseParser.raise_request_error(resp)

        json_data = resp.json()

        if (JsonKeys.SUCCESS in json_data and not json_data[JsonKeys.SUCCESS]) or (
            JsonKeys.ERROR in json_data
        ):
            ResponseParser.raise_request_error(resp)

        statuses = [
            DeviceStatus.from_json({key: json_data[key]})
            for key in json_data
            if key != JsonKeys.SUCCESS
        ]

        return cls(statuses)



"""Tests for the control_device method of the Qwikswitch API client."""

import pytest
import requests

from qwikswitchapi.exceptions import QSRequestError, QSRequestFailedError
from qwikswitchapi.utility import UrlBuilder


def test_success_returns_control_result(authenticated_api_client, mock_request):
    device_id = "@112331"
    level = 50
    response = {"success": True, "device": device_id, "level": level}

    mock_request.get(
        UrlBuilder.build_control_url(
            authenticated_api_client._api_keys.read_write_key, device_id, level
        ),
        json=response,
    )
    result = authenticated_api_client.control_device(device_id, level)

    assert mock_request.called
    assert result is not None
    assert result.device_id == device_id
    assert result.level == level


@pytest.mark.parametrize(
    "response",
    [{"error": "INVALID LEVEL"}, {"error": "INVALID DEVICE ID"}, {"success": False}],
)
def test_logical_error_raises_exception(
    response, authenticated_api_client, mock_request
):
    device = "@112331"
    level = -1
    mock_request.get(
        UrlBuilder.build_control_url(
            authenticated_api_client._api_keys.read_write_key, device, level
        ),
        json=response,
    )

    with pytest.raises(QSRequestError):
        authenticated_api_client.control_device(device, level)


def test_error_raises_exception(authenticated_api_client, mock_request):
    device = "@112331"
    level = -1
    mock_request.get(
        UrlBuilder.build_control_url(
            authenticated_api_client._api_keys.read_write_key, device, level
        ),
        exc=requests.exceptions.Timeout,
    )

    with pytest.raises(QSRequestFailedError):
        authenticated_api_client.control_device(device, level)


def test_with_unknown_error_throws_exception(authenticated_api_client, mock_request):
    device = "@112331"
    level = -1
    mock_request.get(
        UrlBuilder.build_control_url(
            authenticated_api_client._api_keys.read_write_key, device, level
        ),
        status_code=401,
    )

    with pytest.raises(QSRequestError):
        authenticated_api_client.control_device(device, level)



import pytest
import requests_mock

from qwikswitchapi.client import QSClient
from qwikswitchapi.entities import ApiKeys


@pytest.fixture
def mock_request():
    with requests_mock.Mocker() as m:
        yield m


@pytest.fixture
def mock_api_keys():
    return ApiKeys("read_key", "read_write_key")


@pytest.fixture
def api_client():
    return QSClient("email", "master")


@pytest.fixture
def authenticated_api_client(mock_api_keys):
    client = QSClient("email", "master")
    client._api_keys = mock_api_keys
    return client



"""The QwikSwitch API client."""

import functools
from typing import Any

import requests
from requests.exceptions import RequestException

from .constants import DEFAULT_BASE_URI, DEFAULT_TIMEOUT, JsonKeys
from .entities import ApiKeys, ControlResult, DeviceStatuses
from .utility import ResponseParser, UrlBuilder


class QSClient:
    """The QwikSwitch API client."""

    def _ensure_authenticated(func):  # type: ignore # noqa: N805
        @functools.wraps(func)  # type: ignore
        def authenticate_if_needed(self: Any, *args: Any, **kwargs: Any):
            if self._api_keys is None:
                self.generate_api_keys()
            return func(self, *args, **kwargs)  # type: ignore

        return authenticate_if_needed

    def _handle_request_failure(func):  # type: ignore  # noqa: N805
        @functools.wraps(func)  # type: ignore
        def catch_failure(self, *args: Any, **kwargs: Any) -> Any:  # noqa: ANN001
            try:
                return func(self, *args, **kwargs)  # type: ignore
            except RequestException as ex:
                url = ex.request.url if ex.request is not None else "Unknown"
                ResponseParser.raise_request_failure(url, ex)  # type: ignore

        return catch_failure

    def __init__(
        self, email: str, master_key: str, base_uri: str = DEFAULT_BASE_URI
    ) -> None:
        """
        Initialize a new instance of the QSApi class.

        :param email: the email address to generate API keys for:param email: your email address registered on https://qwikswitch.com
        :param master_key: 12 character key found under your CloudHub.  This should be your device id of your Qwikswitch Wi-Fi bridge.
        :param base_uri: the base URI of the Qwikswitch API, optional.  Defaults to 'https://qwikswitch.com/api/v1/'
        """
        self._email = email
        self._master_key = master_key
        self._api_keys = None

        if not base_uri.endswith("/"):
            base_uri += "/"

        self._base_uri = base_uri

    @property
    def base_uri(self) -> str:
        """
        The base URI of the Qwikswitch API.

        :returns: The base URI of the Qwikswitch API
        """
        return self._base_uri

    @property
    def api_keys(self) -> ApiKeys | None:
        """
        The API keys for the QwikSwitch API.

        :returns: The API keys for the QwikSwitch API
        """
        return self._api_keys

    @api_keys.setter
    def api_keys(self, value: ApiKeys) -> None:
        """Set the API keys for the QwikSwitch API."""
        self._api_keys = value

    @_handle_request_failure  # type: ignore
    def generate_api_keys(self) -> ApiKeys:
        """
        Generate API keys for the given email and master key to be used in subsequent calls.

        :returns: APIKeys, with an API key for read operations, and one for read-write operations.
        :raises QSException: on failure to generate API keys
        """
        url = UrlBuilder.build_generate_api_keys_url(self._base_uri)
        req = {JsonKeys.EMAIL: self._email, JsonKeys.MASTER_KEY: self._master_key}

        resp = requests.post(url, json=req, timeout=DEFAULT_TIMEOUT)
        self._api_keys = ApiKeys.from_resp(resp)
        return self._api_keys

    @_handle_request_failure  # type: ignore
    def delete_api_keys(self) -> None:
        """
        Delete API keys generated for the given email and master key.

        :returns: None
        :raises QSException: on failure to delete API keys
        """
        url = UrlBuilder.build_delete_api_keys_url(self._base_uri)
        req = {JsonKeys.EMAIL: self._email, JsonKeys.MASTER_KEY: self._master_key}

        resp = requests.post(url, json=req, timeout=DEFAULT_TIMEOUT)
        _ = ApiKeys.from_resp(resp)

    @_ensure_authenticated  # type: ignore
    @_handle_request_failure  # type: ignore
    def control_device(self, device_id: str, level: int) -> ControlResult:
        """
        Control a device by setting the desired level.

        :param device_id: the unique identifier of the device to control
        :param level: this is a description of what is returned
        :returns: ControlResult, with the device and level set
        :raises QSException: when the request fails
        """
        url = UrlBuilder.build_control_url(
            self._api_keys.read_write_key,  # type: ignore
            device_id,
            level,
            self._base_uri,
        )

        resp = requests.get(url, timeout=DEFAULT_TIMEOUT)
        return ControlResult.from_resp(resp)

    @_ensure_authenticated  # type: ignore
    @_handle_request_failure  # type: ignore
    def get_all_device_status(self) -> DeviceStatuses:
        """
        Retrieve the status of all devices registered to the given API keys.

        :param auth: authentication keys generated by generate_api_keys
        :returns: Array of DeviceStatus with device information
        :raises QSException: when the request fails
        """
        url = UrlBuilder.build_get_all_device_status_url(
            self._api_keys.read_write_key,  # type: ignore
            self._base_uri,
        )

        resp = requests.get(url, timeout=DEFAULT_TIMEOUT)
        return DeviceStatuses.from_resp(resp)

    _ensure_authenticated = staticmethod(_ensure_authenticated)
    _handle_request_failure = staticmethod(_handle_request_failure)



"""Utility methods for handling urls and parsing response messages."""

from typing import Never
from urllib.parse import quote_plus, urljoin

from requests import RequestException

from .constants import DEFAULT_BASE_URI
from .exceptions import (
    QSAuthError,
    QSRequestError,
    QSRequestFailedError,
)


class ResponseParser:
    """Utility methods to parse and validate HTTP responses."""

    @staticmethod
    def get_failure_message(resp) -> str:  # noqa: ANN001
        """
        Return a formatted message indicating the request failed.

        :param resp: The response object
        :return: A formatted message indicating the request failed
        """
        request_url = resp.request.url
        return (
            f'Failed to call {request_url}.  Status code: "{resp.status_code}", '
            f'body: "{resp.text}"'
        )

    @staticmethod
    def raise_request_failure(url, ex: RequestException) -> Never:  # noqa: ANN001
        """
        Raise a QSRequestFailedError indicating the request failed.

        :param url: The URL of the request
        :param ex: The exception that was raised
        :raises QSRequestFailedError: Indicating the request failed, chained with the original exception.
        """
        msg = f"Request to {url} failed: {ex!s}"
        raise QSRequestFailedError(msg) from ex

    @staticmethod
    def raise_request_error(resp) -> Never:  # noqa: ANN001
        """
        Raise a QSRequestError indicating the request failed.

        :param resp: The response object
        :raises QSRequestError: Indicating the request failed, with the body of the response.
        """
        raise QSRequestError(ResponseParser.get_failure_message(resp))

    @staticmethod
    def raise_auth_failure(resp) -> Never:  # noqa: ANN001
        """
        Raise a QSAuthException indicating the request failed.

        :param resp: The response object
        :raises QSAuthError: Indicating the request failed, with the body of the response.
        """
        raise QSAuthError(ResponseParser.get_failure_message(resp))


class UrlBuilder:
    """Utility methods to build URLs for the Qwikswitch API."""

    @staticmethod
    def build_get_all_device_status_url(
        key: str, base_uri: str = DEFAULT_BASE_URI
    ) -> str:
        """
        Build the URL to get the status of all devices.

        :param base_uri: The base URI to use
        :param key: The API key to use
        :return: The url to the 'Get all device status' endpoint
        """
        return urljoin(base_uri, f"state/{quote_plus(key)}/")

    @staticmethod
    def build_generate_api_keys_url(base_uri: str = DEFAULT_BASE_URI) -> str:
        """
        Build the URL to generate API keys.

        :param base_uri: The base URI to use
        :return: The url to the 'Generate API keys' endpoint
        """
        return urljoin(base_uri, "keys")

    @staticmethod
    def build_delete_api_keys_url(base_uri: str = DEFAULT_BASE_URI) -> str:
        """
        Build the URL to delete API keys.

        :param base_uri: The base URI to use
        :return: The url to the 'Delete API keys' endpoint
        """
        return urljoin(base_uri, "keys/delete/")

    @staticmethod
    def build_control_url(
        key: str, device_id: str, level: int, base_uri: str = DEFAULT_BASE_URI
    ) -> str:
        """
        Build the URL to control a device.

        :param base_uri: The base URI to use
        :param key: The API key to use
        :param device_id: The unique identifier of the device to control
        :param level: The level to set the device to
        :return: The url to the `Control a device` endpoint
        """
        return urljoin(
            base_uri,
            f"control/{quote_plus(key)}/"
            f"?device={quote_plus(device_id)}&setlevel={level}",
        )



"""Base Tests for the QSClient class."""

from qwikswitchapi.client import QSClient


def test_base_uri_without_trailing_slash_gets_added():
    api = QSClient("email", "master", base_uri="https://qwikswitch.com/api/v1")
    assert api.base_uri == "https://qwikswitch.com/api/v1/"



"""Tests for the get_all_device_status method of the Qwikswitch API client."""

import pytest
import requests.exceptions

from qwikswitchapi.constants import DeviceClass
from qwikswitchapi.exceptions import QSRequestError, QSRequestFailedError
from qwikswitchapi.utility import UrlBuilder


def test_with_valid_credentials_returns_device_statuses(
    authenticated_api_client, mock_request
):
    response = {
        "success": True,
        "@11111a": {
            "type": "RELAY QS-D-S5",
            "hardware": "0x81",
            "firmware": "v3.3",
            "epoch": "1736018165",
            "rssi": "59%",
            "value": 0,
        },
        "@11111b": {
            "type": "RELAY QS-Q-S9",
            "hardware": "0x81",
            "firmware": "v3.3",
            "epoch": "1736018046",
            "rssi": "58%",
            "value": 0,
        },
    }

    mock_request.get(
        UrlBuilder.build_get_all_device_status_url(
            authenticated_api_client._api_keys.read_write_key
        ),
        json=response,
    )

    devices = authenticated_api_client.get_all_device_status()

    assert mock_request.called
    assert devices is not None
    assert devices.statuses is not None
    assert len(devices.statuses) == 2
    assert devices.statuses[0].device_id == "@11111a"
    assert devices.statuses[1].device_id == "@11111b"

    assert devices.statuses[0].rssi == 59
    assert devices.statuses[1].rssi == 58

    assert devices.statuses[0].device_class == DeviceClass.dimmer
    assert devices.statuses[1].device_class == DeviceClass.unknown


def test_with_unknown_error_throws_exception(authenticated_api_client, mock_request):
    mock_request.get(
        UrlBuilder.build_get_all_device_status_url(
            authenticated_api_client._api_keys.read_write_key
        ),
        status_code=401,
    )

    with pytest.raises(QSRequestError):
        authenticated_api_client.get_all_device_status()


@pytest.mark.parametrize("response", [{"error": "INVALID_API_KEY"}, {"success": False}])
def test_logical_error_raises_exception(
    response, authenticated_api_client, mock_request
):
    mock_request.get(
        UrlBuilder.build_get_all_device_status_url(
            authenticated_api_client._api_keys.read_write_key
        ),
        json=response,
    )

    with pytest.raises(QSRequestError):
        authenticated_api_client.get_all_device_status()


def test_error_raises_exception(authenticated_api_client, mock_request):
    mock_request.get(
        UrlBuilder.build_get_all_device_status_url(
            authenticated_api_client._api_keys.read_write_key
        ),
        exc=requests.exceptions.Timeout,
    )

    with pytest.raises(QSRequestFailedError):
        authenticated_api_client.get_all_device_status()



"""Tests for the generate_api_keys method of the Qwikswitch API client."""

import pytest
import requests.exceptions

from qwikswitchapi.client import QSClient
from qwikswitchapi.entities import ApiKeys
from qwikswitchapi.exceptions import QSAuthError, QSError, QSRequestFailedError
from qwikswitchapi.utility import UrlBuilder


def test_with_valid_credentials_returns_keys(api_client, mock_request):
    response = {"ok": 1, "r": "aaaa-bbbb-cccc-dddd", "rw": "1111-2222-3333-4444"}

    mock_request.post(UrlBuilder.build_generate_api_keys_url(), json=response)
    keys = api_client.generate_api_keys()

    assert mock_request.called
    assert keys is not None
    assert keys.read_key == response["r"]
    assert keys.read_write_key == response["rw"]


def test_with_logical_error_throws_exception(api_client, mock_request):
    response = {
        "ok": 0,
        "err": "Please provide a valid serial key and email address of the registered owner.",
    }

    mock_request.post(UrlBuilder.build_generate_api_keys_url(), json=response)

    with pytest.raises(QSAuthError):
        api_client.generate_api_keys()


def test_with_error_throws_exception(api_client, mock_request):
    mock_request.post(
        UrlBuilder.build_generate_api_keys_url(), exc=requests.exceptions.Timeout
    )

    with pytest.raises(QSRequestFailedError):
        api_client.generate_api_keys()


def test_with_unknown_error_throws_exception(api_client, mock_request):
    response = {"ok": 0}

    mock_request.post(UrlBuilder.build_generate_api_keys_url(), json=response)

    with pytest.raises(QSAuthError):
        api_client.generate_api_keys()


def test_with_invalid_credentials_unknown_error_throws_exception(
    api_client, mock_request
):
    mock_request.post(UrlBuilder.build_generate_api_keys_url(), status_code=401)

    with pytest.raises(QSError):
        api_client.generate_api_keys()


def test_method_is_authenticated_on_request(api_client, mock_request):
    @QSClient._ensure_authenticated  # type: ignore
    def needs_authentication(self) -> None:
        pass

    api_client.needs_authentication = needs_authentication

    response = {"ok": 1, "r": "aaaa-bbbb-cccc-dddd", "rw": "1111-2222-3333-4444"}

    mock_request.post(UrlBuilder.build_generate_api_keys_url(), json=response)
    needs_authentication(api_client)

    assert mock_request.called


def test_method_is_not_authenticated_on_request_uses_supplied_credentials(
    api_client, mock_request
):
    @QSClient._ensure_authenticated  # type: ignore
    def needs_authentication(self) -> None:
        pass

    response = {"ok": 1, "r": "aaaa-bbbb-cccc-dddd", "rw": "1111-2222-3333-4444"}
    mock_request.post(UrlBuilder.build_generate_api_keys_url(), json=response)

    api_client.needs_authentication = needs_authentication
    api_client.api_keys = ApiKeys("read", "read_write")

    needs_authentication(api_client)

    assert not mock_request.called



"""Tests for the delete_api_keys method of the Qwikswitch API client."""

import pytest
import requests

from qwikswitchapi.exceptions import QSAuthError, QSRequestFailedError
from qwikswitchapi.utility import UrlBuilder


def test_with_valid_credentials_returns_none(api_client, mock_request):
    response = {"ok": 1, "r": None, "rw": None}

    mock_request.post(UrlBuilder.build_delete_api_keys_url(), json=response)
    api_client.delete_api_keys()
    assert mock_request.called


def test_with_logical_error_throws_exception(api_client, mock_request):
    response = {
        "ok": 0,
        "err": "Please provide a valid serial key and email address of the registered owner.",
    }

    mock_request.post(UrlBuilder.build_delete_api_keys_url(), json=response)

    with pytest.raises(QSAuthError):
        api_client.delete_api_keys()


def test_with_error_throws_exception(api_client, mock_request):
    mock_request.post(
        UrlBuilder.build_delete_api_keys_url(), exc=requests.exceptions.Timeout
    )

    with pytest.raises(QSRequestFailedError):
        api_client.delete_api_keys()


def test_with_unknown_error_throws_exception(api_client, mock_request):
    response = {"ok": 0}

    mock_request.post(UrlBuilder.build_delete_api_keys_url(), json=response)

    with pytest.raises(QSAuthError):
        api_client.delete_api_keys()


def test_with_invalid_credentials_unknown_error_throws_exception(
    api_client, mock_request
):
    mock_request.post(UrlBuilder.build_delete_api_keys_url(), status_code=401)

    with pytest.raises(QSAuthError):
        api_client.delete_api_keys()



"""Exceptions for the Qwikswitch API."""


class QSError(Exception):
    """
    Base exception raised by the Qwikswitch API.

    Source exceptions are chained.
    """


class QSAuthError(QSError):
    """
    Exception raised by the Qwikswitch API when authentication fails.

    Source exceptions are chained.
    """


class QSRequestFailedError(QSError):
    """
    Exception raised by the Qwikswitch API when a request fails.

    Source exceptions are chained.
    """


class QSRequestError(QSError):
    """
    Exception raised by the Qwikswitch API when a request fails.

    Source exceptions are chained.
    """


class QSResponseParseError(QSRequestError):
    """
    Exception raised by the Qwikswitch API when validations fail on a response.

    Source exceptions are chained.
    """



"""Constants for the Qwikswitch API."""

from enum import Enum
from typing import Final

DEFAULT_BASE_URI: Final = "https://qwikswitch.com/api/v1/"
DEFAULT_TIMEOUT: Final = 10000


class JsonKeys:
    """Constants for names of Json keys."""

    OK: Final = "ok"
    SUCCESS: Final = "success"
    ERROR: Final = "error"
    ERR: Final = "err"
    READ_KEY: Final = "r"
    READ_WRITE_KEY: Final = "rw"
    DEVICE: Final = "device"
    LEVEL: Final = "level"
    RSSI: Final = "rssi"
    TYPE: Final = "type"
    FIRMWARE: Final = "firmware"
    EPOCH: Final = "epoch"
    VALUE: Final = "value"
    EMAIL: Final = "email"
    MASTER_KEY: Final = "masterKey"


class DeviceClass(Enum):
    """Enum for device classes."""

    relay = 1
    dimmer = 2
    humidity_temperature = 3
    unknown = 999


DEVICES = {
    "RELAY QS-D-S5": DeviceClass.dimmer,
    "RELAY QS-R-S5": DeviceClass.relay,
    "RELAY QS-R-S30": DeviceClass.relay,
    # Add more devices here
}
